#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>

#define TSSH_WORD        uint8_t
#define TSSH_DWORD       uint16_t

#define TSSH_WORD_SIZE   8
#define TSSH_WORD_MASK   255


typedef struct {
  unsigned int len;
  TSSH_WORD *data;
} tssh_mpint_t;



#ifdef TSSH_ARCH32
#define TSSH_WORD  uint16_t
#define TSSH_DWORD uint32_t
#define TSSH_WORD_SIZE 16
#define TSSH_WORD_MASK 0xffff
#endif

#ifdef TSSH_ARCH64
#define TSSH_WORD uint32_t
#define TSSH_DWORD uint64_t
#define TSSH_WORD_SIZE 32
#define TSSH_WORD_MASK 0xffffffff
#endif

#define TSSH_MULWORD(A,B,H,L) do { \
   TSSH_DWORD __R= (TSSH_DWORD) (A) * (TSSH_DWORD) (B); \
   (H)= __R >> TSSH_WORD_SHIFT ; \
   (L)= __R; } while(0)


#define TSSH_ADDWORD(A,B,C,R) do {\
   TSSH_DWORD __R= (TSSH_DWORD) (A) + (TSSH_DWORD) (B); \
   (C)= __R >> TSSH_WORD_SHIFT; \
   (R)= __R; } while(0)

#define TSSH_SUBWORD(A,B,C,R) do {\
   TSSH_DWORD __R= (TSSH_DWORD) (A) - (TSSH_DWORD) (B); \
   (C)= __R >> TSSH_WORD_SHIFT; \
   (R)= __R; } while(0)



static tssh_mpint_t *tssh_mpint_zero(int len)
{
  tssh_mpint_t *K;
  K= malloc(sizeof(*K));
  K->data= calloc(len,sizeof(K->data[0]));
  K->len= len;
  return K;
}

static tssh_mpint_t *tssh_mpint_copy(tssh_mpint_t *A)
{
  tssh_mpint_t *K;
  K= malloc(sizeof(*K));
  K->data= malloc(A->len*sizeof(K->data[0]));
  K->len= A->len;
  memcpy(K->data, A->data, sizeof(A->data[0])*A->len);
  return K;
}

static tssh_mpint_free(tssh_mpint_t *K)
{
  free(K->data);
  free(K);
}

static void tssh_mpint_mulword
    (tssh_mpint_t *R, tssh_mpint_t *B, TSSH_DWORD a)
{
  TSSH_DWORD M, Hprev;
  TSSH_WORD *bdata, *bend, *rdata;

  bdata= B->data;
  bend= B->data+B->len;
  rdata= R->data;
  Hprev= 0;

  while(bdata!=bend)
  {
    M= a  *  (TSSH_DWORD) *(bdata++)  + Hprev;
    Hprev= M >> TSSH_WORD_SIZE;
    *(rdata++)= M;
  }
  *rdata= Hprev;
}

static void tssh_mpint_add_shifted
    (tssh_mpint_t *R, tssh_mpint_t *B, int shift)
{
  TSSH_DWORD  C, T;
  TSSH_WORD *bdata, *bend, *rdata;

  bdata= B->data;
  bend= B->data+B->len;
  rdata= R->data + shift;
  C= 0;
  while(bdata!=bend)
  {
    T= (TSSH_DWORD) (*bdata++) + (TSSH_DWORD) (*rdata) + C;
    *(rdata++)= T;
    C= T >> TSSH_WORD_SIZE;
  }
  while(C)
  {
    T= (TSSH_DWORD) (*rdata) + C;
    *(rdata++)= T;
    C= T >> TSSH_WORD_SIZE;
  }
}

tssh_mpint_t* tssh_mpint_mul_simple
    (tssh_mpint_t *A,tssh_mpint_t *B)
{
  tssh_mpint_t *R,*T;
  TSSH_WORD *adata, *aend;
  int aindex;

  R= tssh_mpint_zero(A->len+B->len+1);
  T= tssh_mpint_zero(B->len+1);

  adata= A->data;
  aend= A->data+A->len;
  aindex= 0;

  while(adata!=aend)
  {
    tssh_mpint_mulword(T, B, *adata);
    tssh_mpint_add_shifted(R, T, aindex);
    adata++;
    aindex++;
  }

  tssh_mpint_free(T);
  return R;
}

static void print_bin_simple(char *msg, TSSH_DWORD k)
{
  int i;
  printf("%s", msg);
  for(i=(1<<(TSSH_WORD_SIZE*2-1));i!=0;i>>=1)
  {
    if (k&i) printf("1"); else printf("0");
    if (i==(1<<TSSH_WORD_SIZE)) printf(" ");
  }
  printf("\n");
}

void tssh_mpint_sub_simple(tssh_mpint_t *A,tssh_mpint_t *B)
{
  int i;
  TSSH_DWORD  C,K;
  C= 0;
  for(i=0;i<B->len;i++)
  {
    K= (TSSH_DWORD) A->data[i] - C - (TSSH_DWORD) B->data[i];
    A->data[i]= K; 
    C= (K >> TSSH_WORD_SIZE) & 1;
  }
  while(C)
  {
    K= (TSSH_DWORD) A->data[i] - C;
    A->data[i]= K;
    C= (K >> TSSH_WORD_SIZE) & 1;
    i++;
  }
}

static void tssh_mpint_highest_bit_index
    (tssh_mpint_t *T, int *Rnwords, int *Rbit)
{
  int i;
  int hw;
  TSSH_WORD V;
  for(i=T->len-1;i>0;i--)
    if (T->data[i]) break;
  // hw is highest word index.. it can be zero.
  hw= i;
  V= T->data[i];
  for(i=TSSH_WORD_SIZE-1;i>=0;i--)
    if (V & (1<<i)) break;
  if (i==-1) *Rbit= 0; else { *Rbit= hw*TSSH_WORD_SIZE + i; }
  *Rnwords= hw+1;
}

static int tssh_mpint_greater(tssh_mpint_t *A, tssh_mpint_t *B)
{
  int hibitA, hibitB,  nwA, nwB;
  int i;
  tssh_mpint_highest_bit_index(A, &nwA, &hibitA);
  tssh_mpint_highest_bit_index(B, &nwB, &hibitB);

  if (hibitA < hibitB) return 0;
  if (hibitA > hibitB) return 1;
  for(i=nwA-1;i>=0;i--)
  {
    if (A->data[i] > B->data[i]) return 1;
    if (A->data[i] < B->data[i]) return 0; 
  }
  return 0;
}

static void tssh_mpint_shift_left(tssh_mpint_t *A, int S)
{
  TSSH_DWORD LO, K, mask;
  int nw, hibit;
  int i;
  int words_to_shift, bits_to_shift, space_in_highest_elt, extra_words;
  int nlen; 
  TSSH_WORD *ndata;

  tssh_mpint_highest_bit_index(A, &nw, &hibit);
  words_to_shift= S / TSSH_WORD_SIZE;
  bits_to_shift= S % TSSH_WORD_SIZE;
  space_in_highest_elt= TSSH_WORD_SIZE - (hibit%TSSH_WORD_SIZE)-1;

  if (words_to_shift!=0) goto need_to_shift_words;

  if (space_in_highest_elt>=bits_to_shift || A->len>nw)
     // this shifts stuff according to A->len, so we already
     // have the necessary empty space there for the case
     // where the topmost word would overflow.
     goto do_shift_bits;

  extra_words= 1;
  goto need_alloc_extra;

need_to_shift_words:
  extra_words= words_to_shift;
  if (space_in_highest_elt<bits_to_shift) extra_words++;
  if (A->len<(nw+extra_words)) goto need_alloc_extra;

  // shift words in place.
  for(i=nw-1;i>=0;i--)
    A->data[i+words_to_shift]= A->data[i];
  for(i=0;i<words_to_shift;i++)
    A->data[i]= 0;
  goto do_shift_bits;

need_alloc_extra:
  nlen= nw + extra_words;
  ndata= calloc(nlen, sizeof(ndata[0]));
  for(i=0;i<nw;i++)
    ndata[i+words_to_shift]= A->data[i];
  free(A->data);
  A->data= ndata;
  A->len= nlen;

do_shift_bits:
  S= bits_to_shift;
  mask= (1<<S)-1;
  LO= 0;
  for(i=0;i<A->len;i++)
  {
    K= A->data[i];
    K= (K<<S) | LO;
    A->data[i]= K;
    LO= (K>>TSSH_WORD_SIZE) & mask;
  }
}

static void tssh_mpint_shift_right(tssh_mpint_t* A, int S)
{
  int i;
  TSSH_DWORD K, HI, MASK;

  int words_to_shift, bits_to_shift;

  words_to_shift= S / TSSH_WORD_SIZE;
  bits_to_shift= S % TSSH_WORD_SIZE;

  if (words_to_shift)
  {
    for(i=words_to_shift;i<A->len;i++)
    {
      A->data[i-words_to_shift]= A->data[i];
    }
    A->len -= words_to_shift;
  }

  S= bits_to_shift;

  MASK= ((1<<S)-1) << (TSSH_WORD_SIZE*2-S);
  HI= 0;
  for(i=A->len-1;i>=0;i--)
  {
    K= A->data[i];
    K= K<<TSSH_WORD_SIZE;
    K= (K >> S) | HI;
    A->data[i]= K >> TSSH_WORD_SIZE;
    HI= (K << TSSH_WORD_SIZE) & MASK;
  }
}

static void tssh_mpint_tighten(tssh_mpint_t *A)
{
  int i;
  for(i=A->len-1;i>=0;i--)
   if (A->data[i]) break;
  A->len= i+1;
  if (!A->len) A->len= 1;
}

static int tssh_mpint_impossible_situation1;
static int tssh_mpint_impossible_situation2;

static int tssh_mpint_mulmod_compute_c_k
   (tssh_mpint_t *B, tssh_mpint_t *P, tssh_mpint_t **RC, int *Rk)
{
  int nwordsP, nwordsB, hibitP, hibitB;
  tssh_mpint_t *C;
  int delta;

  C= tssh_mpint_copy(B);
  if (!tssh_mpint_greater(P,B))
  {
    // this is actually impossible, but I will handle it anyway.
    tssh_mpint_impossible_situation1++;
    tssh_mpint_free(C);
    return 1;
  }
  tssh_mpint_highest_bit_index(P, &nwordsP, &hibitP);
  tssh_mpint_highest_bit_index(B, &nwordsB, &hibitB);
  delta= hibitP - hibitB;
  if (delta)
    tssh_mpint_shift_left(C, delta);
  if (tssh_mpint_greater(P,C))
  {
    tssh_mpint_shift_left(C, 1);
    delta++;
  }
  if (tssh_mpint_greater(P,C))
  {
    tssh_mpint_impossible_situation2++;
    tssh_mpint_free(C);
    return 1;
  }
  tssh_mpint_sub_simple(C,P);
  *Rk= delta;
  *RC= C;
printf("COMPUTE C, K ====\n");
print_number("P= ", P);
print_number("B= ", B);
print_number("C= ", C);
printf("k= %d\n", delta);
printf("^^^^^^^^^^^^^^^^^^\n");
  return 0;
}

static void tssh_mpint_remainder(tssh_mpint_t *A, tssh_mpint_t *P)
{
  while(tssh_mpint_greater(A,P)) tssh_mpint_sub_simple(A,P);
  if (tssh_mpint_greater(P,A)) return ;
  tssh_mpint_sub_simple(A,P);
}

static tssh_mpint_t *tssh_mpint_lowpart(tssh_mpint_t *A, int nbits)
{
  int nwords;
  tssh_mpint_t *R;
  TSSH_WORD mask;
  int i;

  nwords= (nbits+TSSH_WORD_SIZE-1)/TSSH_WORD_SIZE;
  R= tssh_mpint_zero(nwords);
  for(i=0;i<nwords;i++) R->data[i]= A->data[i];

  nbits %= TSSH_WORD_SIZE;
  if (nbits)
  {
    mask= (1<<nbits)-1;
    R->data[nwords-1] &= mask;
  }
  return R;
}

void tssh_mpint_mulmod_PROC_mul_to_accu
    (tssh_mpint_t *accu, tssh_mpint_t *A, tssh_mpint_t *B)
{
  tssh_mpint_t *T;
  TSSH_WORD *adata, *aend;
  int aindex;

  T= tssh_mpint_zero(B->len+1);

  adata= A->data;
  aend= A->data+A->len;
  aindex= 0;

  while(adata!=aend)
  {
    tssh_mpint_mulword(T, B, *adata);
    tssh_mpint_add_shifted(accu, T, aindex);
    adata++;
    aindex++;
  }

  tssh_mpint_free(T);
}

static tssh_mpint_t *tssh_mpint_mulmod
   (tssh_mpint_t *A, tssh_mpint_t *B, tssh_mpint_t *P)
{
   tssh_mpint_t *accumulator;
   tssh_mpint_t *C, *AL;
   int k;
   int nwordsA, hibitA;

    // I just put a +1 word in case somebody wants to write
    // zeroes there
   accumulator= tssh_mpint_zero(P->len+2);

   A= tssh_mpint_copy(A);
   B= tssh_mpint_copy(B);

restart:
   tssh_mpint_tighten(A); if (A->len==1 && A->data[0]==0) goto done;
   tssh_mpint_tighten(B); if (B->len==1 && B->data[0]==0) goto done;
#if 1
   if (tssh_mpint_greater(B,A)) 
   {
     tssh_mpint_t *t;
     t= B;
     B= A;
     A= t;
   }
#endif
printf("MULMOD ROUND BEGIN =======\n");
print_number("iA= ",A);
print_number("iB= ",B);
print_number("P= ",P);
print_number("total= ", accumulator);

   tssh_mpint_highest_bit_index(A, &nwordsA, &hibitA);
   if (tssh_mpint_mulmod_compute_c_k(B, P, &C, &k)) goto fail;
   if (k>hibitA) goto nopart;

printf("Split at  bit %d\n", k);

   AL= tssh_mpint_lowpart(A, k);
   tssh_mpint_shift_right(A, k);

print_number("AL= ", AL);
print_number("A= ", A);

   tssh_mpint_mulmod_PROC_mul_to_accu(accumulator, AL, B);
print_number("Accu (After mul) ", accumulator);
   tssh_mpint_remainder(accumulator, P);
print_number("Accu (After mod) ", accumulator);

   tssh_mpint_free(B);
   tssh_mpint_free(AL);
   B= C;
   goto restart;
fail:
   printf("computation of C and k failed\n");
   *(int*) NULL= 0;
   return ;
nopart:
  /* there was no partition possible because the first k for
     which B*2^k>P was greater than the highest bit of A.
     Therefore, AH would be zero. Here we simply do the
     multiplication with A and call it a day. */
printf("NO NEED FOR PARTITION MUL A*B\n");
print_number("A= ",A);
print_number("B= ",B);
print_number("accu (pre-mul)= ",accumulator);
   tssh_mpint_mulmod_PROC_mul_to_accu(accumulator, A, B);
done:
print_number("accu (before remainder)= ",accumulator);
   tssh_mpint_remainder(accumulator, P);
print_number("accu (final)= ",accumulator);
   tssh_mpint_free(A);
   tssh_mpint_free(B);
   return accumulator;
}


static uint32_t mulmod_uint32(uint32_t a,uint32_t b,uint32_t p)
{
  uint64_t A,B,P;
  A= a; B= b; P= p;
  return (A*B)%P;
}


static uint64_t collect_number(tssh_mpint_t *num)
{
  int i,shift;
  uint64_t V,K;
  V= 0;
  shift= 0;
  for(i=0;i<num->len;i++)
  {
    K= num->data[i];
    K<<= shift;
    V|= K;
    shift+= TSSH_WORD_SIZE;
  }
  return V;
}

static void print_binary(TSSH_WORD V)
{
  TSSH_WORD K;
  K= 1<<(TSSH_WORD_SIZE-1);
  while(K)
  {
     if (V&K) printf("1"); else printf("0");
     K>>= 1;
  }
}

void print_number(char *msg,tssh_mpint_t *num)
{
  int i,L;
  int digitslen;
  char buf[200];
  printf("%s", msg);
  L= strlen(msg);
  digitslen= num->len*(TSSH_WORD_SIZE+1);
  sprintf(buf,"(%d)", collect_number(num));
  printf("%s",buf);
  L+= strlen(buf);
  for(i=L+digitslen;i<79;i++) printf(" ");
  for(i=num->len-1;i>=0;i--)
  {
    print_binary(num->data[i]);
    printf(" ");
  }
  printf("\n");
}

static uint64_t safe_op(uint32_t a, uint32_t b, int op)
{
  uint64_t A, B;
  A= a;
  B= b;
  switch(op)
  {
   case '+': return A+B;
   case '-': return A-B;
   case '*': return A*B;
  }
  return A;
}

tssh_mpint_t *tssh_mpint_from_uint32(uint32_t V)
{
  tssh_mpint_t *T;
  int i,maxlen,mask,shift;
  maxlen = 32 / TSSH_WORD_SIZE;
  for(i=maxlen;i>1;i--)
  {
    if (V >> ((i-1)*TSSH_WORD_SIZE))
      break;
  }
  T= tssh_mpint_zero( i );
  i= 0;
  while(V)
  {
    T->data[i]= V & TSSH_WORD_MASK;
    V>>=TSSH_WORD_SIZE;
    i++;
  }
  return T;
}

static int get_small_number(char *msg)
{
  char str[2000];
  printf("%s:",msg); fflush(stdout);
  memset(str,0,sizeof(str));
  fgets(str, sizeof(str), stdin);
  str[sizeof(str)-1]= 0;
  return atoi(str);
}

static tssh_mpint_t *get_number(char *msg)
{
  char str[2000];
  int digits, i;
  int bit, byte;
  tssh_mpint_t *num;

  printf("%s:",msg); fflush(stdout);
  memset(str,0,sizeof(str));
  fgets(str, sizeof(str), stdin);
  str[sizeof(str)-1]= 0;

  if (str[0]!='b') return tssh_mpint_from_uint32(strtoul(str,NULL,0));
  digits= 0;
  for(i=0;str[i];i++) if (str[i]=='0' || str[i]=='1') digits++;

  num= tssh_mpint_zero( (digits + (TSSH_WORD_SIZE-1))/(TSSH_WORD_SIZE) );
  byte= 0;
  bit= 1;
  for(i=strlen(str)-1;i>=0;i--)
  {
    if (str[i]=='1') num->data[byte] |= bit;
    if (str[i]=='0' || str[i]=='1')
    {
     bit= bit<<1;
     if (bit>=(1<<TSSH_WORD_SIZE)) { bit= 1; byte++; }
    }
  }
  return num;
}

uint32_t make_random_uint32()
{
  int i;
  uint8_t K[4];
  for(i=0;i<4;i++) K[i]= rand();
  return K[0] | (K[1]<<8) | (K[2]<<16) | (K[3]<<24);
}

void test(int N,int op)
{
  int i;
  uint32_t a,b;
  tssh_mpint_t *A,*B,*R;

  srand(time(NULL));
  for(i=0;i<N;i++)
  {
    b= make_random_uint32();
    while(!(a= make_random_uint32()));
    b %= a;
    A= tssh_mpint_from_uint32(a);
    B= tssh_mpint_from_uint32(b);
    switch(op)
    {
    case '-': tssh_mpint_sub_simple(A,B); R= A; break;
    case '*': R= tssh_mpint_mul_simple(A,B); break;
    }
    if (collect_number(R) != safe_op(a,b,op))
    {
      printf("fail: %u %c %u \n",a,op,b);
      printf("result should be %llu \n", safe_op(a,b,op));
      printf("        achieved %llu \n", collect_number(R));
      exit(1);
    }
  }
}

void Tsub(uint64_t v1, uint64_t v2)
{
  tssh_mpint_t *A,*B;
  A= tssh_mpint_from_uint32(v1);
  B= tssh_mpint_from_uint32(v2);
  print_number("A= ", A);
  print_number("B= ", B);
  tssh_mpint_sub_simple(A,B);
  print_number("R= A-B= ", A);
  if (collect_number(A)!=(v1-v2))
  {
    printf("failed: subtraction\n");
    *(int*) NULL= 0;
  }
}



typedef struct
{
  int len;
  TSSH_WORD *T, *B, *P;
  TSSH_WORD *data;
} tssh_MULMOD_t;


static void tssh_MULMOD_addmod(tssh_MULMOD_t *MM, TSSH_WORD *G)
{
  TSSH_WORD *B,*P;
  TSSH_DWORD K, C;
  int L,i;
  L= MM->len;
  B= MM->B;

  C= 0;
  for(i=0;i<L;i++)
  {
    K= (TSSH_DWORD) G[i] + (TSSH_DWORD) B[i] + C;
    G[i]= K;
    C= (K>>TSSH_WORD_SIZE);
  }

  P= MM->P;
  for(i=L-1;i>=0;i--)
    if (G[i]>P[i]) break;
    else if (G[i]<P[i]) return ;

  C= 0;
  for(i=0;i<L;i++)
  {
    K= (TSSH_DWORD) G[i] - (TSSH_DWORD) P[i] - C;
    G[i]= K;
    C= (K>>TSSH_WORD_SIZE) & 1;
  }
}

static tssh_MULMOD_t *tssh_MULMOD_begin(tssh_mpint_t *P)
{
  int i;
  tssh_MULMOD_t *MM;
  MM= malloc(sizeof(*MM));
  MM->len= P->len+1;
  MM->data= calloc(MM->len*3, sizeof(MM->data[0]));
  MM->T= MM->data;
  MM->B= MM->T + MM->len;
  MM->P= MM->B + MM->len;
  for(i=0;i<P->len;i++)
    MM->P[i]= P->data[i];
  return MM;
}

static void tssh_MULMOD_do(tssh_MULMOD_t *MM, TSSH_WORD *A, TSSH_WORD *B)
{
  int i,L;
  int count_nonzero_b;
  TSSH_WORD Av;
  int bit;
  TSSH_WORD *Aend;

  L= MM->len;
  count_nonzero_b= 0;
  for(i=0;i<L;i++)
  {
    MM->B[i]= B[i];
    MM->T[i]= 0;
    if (B[i]) count_nonzero_b++;
  }
  if (!count_nonzero_b) return ;

  for(i=L-1;i>=0;i--)
    if (A[i]) break;
  if (i==-1)  // A zero
    return ;

//tssh_mpint_t NNB, NNT, NNP; NNB.len= MM->len; NNB.data= MM->B;
//NNT.len= MM->len; NNT.data= MM->T; NNP.len= MM->len; NNP.data= MM->P;
  Aend= A + i + 1;
next_word:
  Av= A[0];
  bit= 0;
next_bit:
//printf("MULMOD new round===\n"); print_number("B= ", &NNB);
//print_number("T= ", &NNT); print_number("P= ", &NNP);
  if (Av&1) tssh_MULMOD_addmod(MM, MM->T);
  tssh_MULMOD_addmod(MM, MM->B);
  Av >>= 1;
  bit++;
  if (bit!=TSSH_WORD_SIZE) goto next_bit;
  A++;
  if (A!=Aend) goto next_word;
}

static void tssh_MULMOD_end(tssh_MULMOD_t *MM)
{
  memset(MM->data, 0, MM->len*3*sizeof(MM->data[0]));
  free(MM->data);
  free(MM);
}

static tssh_mpint_t* tssh_MULMOD
   (tssh_mpint_t *A, tssh_mpint_t *B,tssh_mpint_t *P)
{
  TSSH_WORD *BD, *AD;
  tssh_mpint_t *R;
  tssh_MULMOD_t *MM;
  int i;

  BD= calloc(P->len+1, sizeof(BD[0]));
  AD= calloc(P->len+1, sizeof(AD[0]));

  for(i=0;i<B->len;i++) BD[i]= B->data[i];
  for(i=0;i<A->len;i++) AD[i]= A->data[i];

  MM= tssh_MULMOD_begin(P);
  tssh_MULMOD_do(MM,AD,BD);
  R= tssh_mpint_zero(P->len);
  memcpy(R->data, MM->T, P->len*sizeof(R->data[0]));
  tssh_MULMOD_end(MM);

  memset(BD, 0, (P->len+1)*sizeof(BD[0]));
  memset(AD, 0, (P->len+1)*sizeof(AD[0]));
  free(BD);
  free(AD);
  return R;
}


void lonum_subtract(TSSH_WORD *A, TSSH_WORD *B, int L)
{
  int i;
  TSSH_DWORD C, K;
  C= 0;
  for(i=0;i<L;i++)
  {
     K= (TSSH_DWORD) A[i] - C - (TSSH_DWORD) B[i];
     A[i]= K;
     C= (K >> TSSH_WORD_SIZE) &1;
  }
}

void lonum_lshift1(TSSH_WORD *dst, TSSH_WORD *src, int L)
{
  TSSH_DWORD C, K;
  int i;
  C= 0;
  for(i=0;i<L;i++)
  {
     K= src[i];
     K= (K<<1) | C;
     dst[i]= K;
     C= (K>>TSSH_WORD_SIZE)&1;
  }
}

void lonum_add(TSSH_WORD *R, TSSH_WORD *T, int L)
{
  TSSH_DWORD C, K;
  int i;
  C= 0;
  for(i=0;i<L;i++)
  {
    K= (TSSH_DWORD) T[i] + C + (TSSH_DWORD) R[i];
    R[i] = K;
    C= K >> TSSH_WORD_SIZE;
  } 
}

void lonum_print(char *msg, TSSH_WORD *data, int len)
{
  tssh_mpint_t N;
  N.len= len;
  N.data= data;
  print_number(msg, &N);
}

typedef struct
{
  TSSH_WORD *multiples[TSSH_WORD_SIZE+1];
  TSSH_WORD *data;
  int len;
} XMOD_t;

XMOD_t *make_Xmod(tssh_mpint_t *P)
{
  XMOD_t *ms;
  int i;

  ms= malloc(sizeof(*ms));
  ms->len= P->len+1;
  ms->data= calloc(ms->len*(TSSH_WORD_SIZE+1),TSSH_WORD_SIZE);

  for(i=0;i<=TSSH_WORD_SIZE;i++)
    ms->multiples[i]= ms->data+i*ms->len;

  for(i=0;i<P->len;i++)
    ms->multiples[0][i]= P->data[i];

  for(i=1;i<=TSSH_WORD_SIZE;i++)
    lonum_lshift1(ms->multiples[i], ms->multiples[i-1], ms->len);

  for(i=0;i<=TSSH_WORD_SIZE;i++)
  {
    char buf[200]; 
    sprintf(buf, "P*2^%d= ", i);
    lonum_print(buf, ms->multiples[i], ms->len);
  }
  
  return ms;
}

void free_Xmod(XMOD_t *ms)
{
  memset(ms->data,0, ms->len*(TSSH_WORD_SIZE+1)*sizeof(ms->data[0]));
  free(ms->data);
  free(ms);
}

static int lonum_greater(TSSH_WORD* A, TSSH_WORD *B, int L)
{
  int i;
  for(i=L-1;i>=0;i--)
    if (A[i]<B[i]) return 0;
    else if (A[i]>B[i]) return 1;
  return 0;
}

void xmod_reduce(XMOD_t *ms, TSSH_WORD *A)
{
  int i;
  for(i=TSSH_WORD_SIZE;i>=0;i--)
    if (!lonum_greater(ms->multiples[i], A, ms->len))
      lonum_subtract(A, ms->multiples[i], ms->len);
}

tssh_mpint_t* mulmod_w(tssh_mpint_t *A, tssh_mpint_t *B, tssh_mpint_t *P)
{
  XMOD_t *ms;
  tssh_mpint_t *T, *R;
  TSSH_DWORD wA, K, C, Hprev;
  TSSH_WORD *bb;
  int iA, i, L;

  ms= make_Xmod(P);
  T= tssh_mpint_zero(P->len+1);
  R= tssh_mpint_zero(P->len+1);
  bb= calloc(P->len+1, sizeof(bb[0]));

  L= ms->len;

  for(i=0;i<B->len;i++) bb[i]= B->data[i];

  iA= 0;
next_digit:
  wA= A->data[iA];

printf("NEW ROUND ===== \n");
lonum_print("B= ", bb, L);
printf("wA= %x\n", wA);

  memset(T->data, 0, sizeof(T->data[0])*T->len);

  Hprev= 0;
  for(i=0;i<L;i++)
  {
    K= (TSSH_DWORD) bb[i] * wA + Hprev;
    T->data[i]= K;
    Hprev= K >> TSSH_WORD_SIZE;
  }
  T->data[i]= Hprev;

lonum_print("T= ",T->data,L);
  xmod_reduce(ms, T->data);
lonum_print("T(reduced)= ",T->data,L);

lonum_print("R(preadd)= ", R->data, L);
  lonum_add(R->data, T->data, L);
lonum_print("R(postadd)= ", R->data, L);
  xmod_reduce(ms, R->data);
lonum_print("R(postreduce)= ", R->data, L);

  iA++;
  if (iA>=A->len) return R;
  
lonum_print("B(preshift-w)", bb, L);
  for(i=L-1;i>0;i--) bb[i]= bb[i-1];
  bb[0]= 0;
lonum_print("B(postshift-w)", bb, L);
  xmod_reduce(ms, bb);
lonum_print("B(postreduce-1)", bb, L);
  goto next_digit;
}




void Tmulmod(uint32_t a, uint32_t b, uint32_t p)
{
  tssh_mpint_t *A, *B, *R;
  tssh_mpint_t *P;

  A= tssh_mpint_from_uint32(a);
  B= tssh_mpint_from_uint32(b);
  P= tssh_mpint_from_uint32(p);
  print_number("A= ",A);
  print_number("B= ",B);
  print_number("P= ",P);
  R= mulmod_w(A,B,P);
  print_number("R= ",R);

  printf("by uint64 R= %d\n", mulmod_uint32(a,b,p));
}

int main(int argc,char **argv)
{
  tssh_mpint_t *A, *B, *R;
  tssh_mpint_t *P;
  int N;
  int nw, bits;
  if (argc>2) { N= atoi(argv[1]); test(N,argv[2][0]);
                 printf("all ok\n"); return 0; }

  //Tmulmod(700,500, 21143);
  // Tmulmod(6990069,1922535,13832537);
  Tmulmod( 9765413, 11214561, 13832537);
  // Tmulmod(get_small_number("num2"),get_small_number("num1"),13832537);
  //Tmulmod(7561034, 12981021,13832939);
  return 0;
#if 0
  // Tsub(15380280, 13832537);

  Tmulmod(get_small_number("num2"),get_small_number("num1"),13832537);
  return 0;

  A= get_number("num1");
  print_number("A= ",A);
  N= get_small_number("shift");
  tssh_mpint_shift_left(A, N);
  print_number("R= ",A);
  return 0;
#endif

  A= get_number("num1");
  print_number("A= ",A);
  tssh_mpint_highest_bit_index(A, &nw, &bits);
  printf("highest bit= %d, nwords= %d\n", bits, nw);
  return 0;

  A= get_number("num1");
  B= get_number("num2");
  print_number("A= ",A);
  print_number("B= ",B);
  tssh_mpint_sub_simple(A,B);
  print_number("R= A.B= ",A);
  return 0;
}

/******* some primes to play with 
13832227 13832239 13832243 13832251 13832263
13832267 13832279 13832281 13832309 13832317
13832327 13832341 13832387 13832389 13832393
13832419 13832473 13832501 13832519 13832521
13832537 13832569 13832579 13832587 13832641
13832701 13832719 13832737 13832743 13832759
13832761 13832771 13832773 13832803 13832807
13832809 13832831 13832869 13832873 13832881
13832939 13832953
*/

