
#ifdef TSSH_ARCH32
#define TSSH_WORD  uint16_t
#define TSSH_DWORD uint32_t
#define TSSH_WORD_SIZE 16
#define TSSH_WORD_MASK 0xffff
#endif

#ifdef TSSH_ARCH64
#define TSSH_WORD uint32_t
#define TSSH_DWORD uint64_t
#define TSSH_WORD_SIZE 32
#define TSSH_WORD_MASK 0xffffffff
#endif

static uint32_t tssh_get_uint32(unsigned char*);

typedef struct {
  unsigned int len;
  TSSH_WORD *data;
} tssh_mpint_t;

static void tssh_mpint_tighten(tssh_mpint_t *A)
{
  int i;
  for(i=A->len-1;i>=0;i--)
   if (A->data[i]) break;
  A->len= i+1;
  if (!A->len) A->len= 1;
}         

static int tssh_nwords_for_bits(int nbits)
{
  return (nbits+TSSH_WORD_SIZE-1)/TSSH_WORD_SIZE;
}

static tssh_mpint_t *tssh_mpint_zero(int len)
{
  tssh_mpint_t *K;
  K= malloc(sizeof(*K));
  K->data= calloc(len,sizeof(K->data[0]));
  K->len= len;
  return K;
}

static tssh_mpint_t *tssh_mpint_copy(tssh_mpint_t *A)
{
  tssh_mpint_t *K;
  K= malloc(sizeof(*K));
  K->data= malloc(A->len*sizeof(K->data[0]));
  K->len= A->len;
  memcpy(K->data, A->data, sizeof(A->data[0])*A->len);
  return K;
}

static tssh_mpint_free(tssh_mpint_t *K)
{
  free(K->data);
  free(K);
}

static tssh_mpint_t *tssh_mpint_from_uint32(uint32_t _K)
{
  tssh_mpint_t *R;
  TSSH_WORD *D;
  TSSH_DWORD K; // let's suppress the gcc warning
  int i;
  K= _K;
  R= tssh_mpint_zero( tssh_nwords_for_bits(32) );
  D= R->data;
  for(i=0;i<32;i+=TSSH_WORD_SIZE)
  {
    *D= K;
    K>>= TSSH_WORD_SIZE;
    D++;
  }
  return R;
}

static int tssh_hexchr_value(int K)
{
  if (K>='0' && K<='9') return K-'0';
  if (K>='a' && K<='f') return K-'a'+10;
  if (K>='A' && K<='F') return K-'A'+10;
  return -1;
}

static tssh_mpint_t *tssh_mpint_from_hexstr(const char *Z)
{
  int ndigits;
  const char *P;
  tssh_mpint_t *R;
  int nbits, shift,i;
  TSSH_WORD *D, W, V;

  ndigits= 0;
  for(i=0;Z[i];i++) if (tssh_hexchr_value(Z[i])!=-1) ndigits++;
  nbits= ndigits*4;
  R= tssh_mpint_zero( tssh_nwords_for_bits(nbits) );

  P= Z+strlen(Z)-1;
  Z--;
  W= 0;
  shift= 0;
  D= R->data;
  while(P!=Z)
  {
    V= tssh_hexchr_value(*P);
    if (V== (TSSH_WORD) -1) { P--; continue; }
    W |=  (V<<shift);
    shift+= 4;
    if (shift==TSSH_WORD_SIZE) { *(D++)= W; W= 0; shift= 0; }
    P--;
  }
  if (shift) *(D++)= W;
  return R;
}

static tssh_mpint_t *tssh_mpint_from_rfc_mpint(uint8_t *data)
{
  int nbits, shift,i;
  tssh_mpint_t *R;
  TSSH_WORD W, V, *D;
  uint32_t len;

  len= tssh_get_uint32(data);
  data+= 4;

  // rfc says that there is no data if length==0. the number is just 0
  if (!len) return tssh_mpint_zero(1);

  nbits= len*8;
  while(data[0]==0 && nbits>0) { data++; nbits-= 8; len--; }

  // rfc says that this is illegal. a zero must not be
  // encoded as a series of zeroes, but I'll allow it for now.
  if (nbits<=0) return tssh_mpint_zero(1);

  R= tssh_mpint_zero(tssh_nwords_for_bits(nbits));
  D= R->data; shift= 0; W= 0;
  for(i=len-1;i>=0;i--)
  {
    V= data[i];
    W |= V << shift;
    shift+= 8;
    if (shift==TSSH_WORD_SIZE)
    {
      shift= 0;
      *(D++)= W;
      W= 0;
    }
  }
  if (shift) *(D++)= W;
  return R;
}

static void tssh_mpint_to_rfc_mpint(tssh_mpint_t *T, uint32_t *L,uint8_t **RD)
{
  int i,j, LWbytes, extra_zero;
  TSSH_WORD LW, W, mask;
  uint8_t *R,*D;
  uint32_t nbytes;

  tssh_mpint_tighten(T);
  if (T->len==1 && T->data[0]==0)
  {
    *RD= calloc(1,4);
    *L= 4;
    return ;
  }
  LW= T->data[T->len-1];
  LWbytes= TSSH_WORD_SIZE / 8;

  mask= 0xff << (TSSH_WORD_SIZE-8);
  for(i=0;i<TSSH_WORD_SIZE/8;i++)
    if (LW&mask) break;
            else { LWbytes--; mask >>= 8; }
  if (LW & (1<<(LWbytes*8-1))) extra_zero= 1;
                         else extra_zero= 0;

  nbytes= (T->len-1)*TSSH_WORD_SIZE/8 + LWbytes + extra_zero;
  R= malloc( nbytes + 4 );
  R[0]= nbytes >> 24;
  R[1]= nbytes >> 16;
  R[2]= nbytes >> 8;
  R[3]= nbytes;
  D= R+4 + nbytes-1;
  for(i=0;i<T->len-1;i++)
  {
    W= T->data[i];
    for(j=0;j<TSSH_WORD_SIZE/8;j++)
    {
       *(D--)= W;
       W >>= 8;
    }
  }
  W= T->data[T->len-1];
  for(j=0;j<LWbytes;j++)
  {
     *(D--)= W;
      W >>= 8;
  }
  if (extra_zero) *(D--)= 0;

  *RD= R;
  *L= nbytes + 4;
}


static void tssh_lonum_subtract(TSSH_WORD *A, TSSH_WORD *B, int L)
{
  int i;
  TSSH_DWORD C, K;
  C= 0;
  for(i=0;i<L;i++)
  {
     K= (TSSH_DWORD) A[i] - C - (TSSH_DWORD) B[i];
     A[i]= K;
     C= (K >> TSSH_WORD_SIZE) &1;
  }
}

static void tssh_lonum_lshift1(TSSH_WORD *dst, TSSH_WORD *src, int L)
{
  TSSH_DWORD C, K;
  int i;
  C= 0;
  for(i=0;i<L;i++)
  {
     K= src[i];
     K= (K<<1) | C;
     dst[i]= K;
     C= (K>>TSSH_WORD_SIZE)&1;
  }
}

static void tssh_lonum_add(TSSH_WORD *R, TSSH_WORD *T, int L)
{
  TSSH_DWORD C, K;
  int i;
  C= 0;
  for(i=0;i<L;i++)
  {
    K= (TSSH_DWORD) T[i] + C + (TSSH_DWORD) R[i];
    R[i] = K;
    C= K >> TSSH_WORD_SIZE;
  } 
}


typedef struct
{
  int len;
  TSSH_WORD **number;
  TSSH_WORD *data;
  int n_number;
} tssh_numbers_t;

static tssh_numbers_t *tssh_numbers_new(int n_number,int len)
{
  tssh_numbers_t *NN;
  int i;
  NN= malloc(sizeof(*NN));
  NN->len= len;
  NN->n_number= n_number;
  NN->data= calloc(n_number*len, sizeof(NN->data[0]));
  NN->number= malloc(sizeof(NN->number[0])*n_number);
  for(i=0;i<n_number;i++)
    NN->number[i]= NN->data+i*len;
  return NN;
}

static tssh_numbers_t *tssh_multiples_of_modulus(TSSH_WORD *P, int len)
{
  tssh_numbers_t *M;
  int i;

  M= tssh_numbers_new(TSSH_WORD_SIZE+1, len+1);
  for(i=0;i<len;i++)
    M->number[0][i]= P[i];

  for(i=1;i<=TSSH_WORD_SIZE;i++)
    tssh_lonum_lshift1(M->number[i], M->number[i-1], len+1);

#if 0
  for(i=0;i<=TSSH_WORD_SIZE;i++)
  {
    char buf[200]; 
    sprintf(buf, "P*2^%d= ", i);
    lonum_print(buf, M->number[i], M->len);
  }
#endif 

  return M;
}

static void tssh_numbers_free(tssh_numbers_t *M) 
{
  memset(M->data,0, M->len*M->n_number*sizeof(M->data[0]));
  free(M->data);
  free(M);
}

static int tssh_lonum_greater(TSSH_WORD* A, TSSH_WORD *B, int L)
{
  int i;
  for(i=L-1;i>=0;i--)
    if (A[i]<B[i]) return 0;
    else if (A[i]>B[i]) return 1;
  return 0;
}

static void tssh_remainder_by_table(TSSH_WORD *A, tssh_numbers_t *MM)
{
  int i;
  for(i=TSSH_WORD_SIZE;i>=0;i--)
    if (!tssh_lonum_greater(MM->number[i], A, MM->len))
      tssh_lonum_subtract(A, MM->number[i], MM->len);
}

static void tssh_mulmod_by_table
   (TSSH_WORD *R, 
    TSSH_WORD *A, int AL, 
    TSSH_WORD *B,
    tssh_numbers_t *modtab,
    TSSH_WORD *temp1, TSSH_WORD *temp2)
{
  TSSH_WORD *T;
  TSSH_DWORD wA, K, C, Hprev;
  TSSH_WORD *bb;
  int iA, i, L;

  L= modtab->len;
  T= temp1;
  bb= temp2;
  for(i=0;i<L;i++) bb[i]= B[i];
  memset(R, 0, sizeof(R[0])*L);

  iA= 0;
next_digit:
  wA= A[iA];

//printf("NEW ROUND\n");
//lonum_print("B= ", bb, L);
//printf("a= %x\n", wA);

  if (!wA) goto skip_multiply;

  memset(T, 0, sizeof(T[0])*L);

  Hprev= 0;
  for(i=0;i<L-1;i++)
  {
    K= (TSSH_DWORD) bb[i] * wA + Hprev;
    T[i]= K;
    Hprev= K >> TSSH_WORD_SIZE;
  }
  T[i]= Hprev;
//lonum_print("T= ", T, L);

  tssh_remainder_by_table(T, modtab);
//lonum_print("T(rem)= ", T, L);
//lonum_print("R= ", R, L);
  tssh_lonum_add(R, T, L);
//lonum_print("R+= ", R, L);
  tssh_remainder_by_table(R, modtab);
//lonum_print("R%= ", R, L);

skip_multiply:
  iA++;
  if (iA>=AL) return ;
  
//lonum_print("PRE<B= ", bb, L);
  for(i=L-1;i>0;i--) bb[i]= bb[i-1];
  bb[0]= 0;
//lonum_print("B<= ", bb, L);
  tssh_remainder_by_table(bb, modtab);
//lonum_print("B%= ", bb, L);
  goto next_digit;
}



static TSSH_WORD *tssh_alloc_number(int L)
{
  return calloc(L,sizeof(TSSH_WORD));
}

static tssh_mpint_t *tssh_expmod_by_table
    (tssh_mpint_t *base,tssh_mpint_t *A, tssh_numbers_t *modtab)
{
  tssh_numbers_t *NN;
  tssh_mpint_t *R;
  int L, bit, Ai,i, H;
  TSSH_WORD Aw;
  TSSH_WORD *B, *Bp, *S, *Sp, 
             *temp1,*temp2, *TT;

  L= modtab->len;

  B= tssh_alloc_number(L);
  Bp= tssh_alloc_number(L);
  S= tssh_alloc_number(L);
  Sp= tssh_alloc_number(L);
  temp1= tssh_alloc_number(L);
  temp2= tssh_alloc_number(L);
  
  for(i=0;i<base->len;i++) B[i]= base->data[i];
  S[0]= 1;

  bit= TSSH_WORD_SIZE-1;
  Ai= -1;
next_bit:
  bit++;
  if (bit==TSSH_WORD_SIZE) 
  {
    Ai++; if (Ai==A->len) goto done;
    Aw= A->data[Ai];
    bit= 0;
  }
  H= Aw & 1; Aw >>= 1;

  if (H)
  {
     tssh_mulmod_by_table(Sp, S, L, B, modtab, temp1, temp2);
     TT= S; S= Sp; Sp= TT;
  }

  tssh_mulmod_by_table(Bp, B, L, B, modtab, temp1, temp2);
  TT= B; B= Bp; Bp= TT;
  goto next_bit;
done:
  R= malloc(sizeof(*R));
  R->len= L;
  R->data= S;
  return R;
}

static int tssh_mpint_greater(tssh_mpint_t *A,tssh_mpint_t *B)
{
  int i;
  tssh_mpint_tighten(A);
  tssh_mpint_tighten(B);
  if (A->len<B->len) return 0;
  if (A->len>B->len) return 1;
  for(i=A->len-1;i>=0;i--)
    if (A->data[i] > B->data[i]) return 1;
  return 0;
}

