/***********************************************************************
 *                                                                     *
 *                             TSSH RAW IO                             *
 *                                                                     *
 ***********************************************************************
  This code deals with sending/receiving things on the socket connected
  to the server. 

  Normally, length of each message has to be a multiple of 8 or
  encryption block size, whichever is greater. However, during the 
  initial exchange, we have an identification string which isn't sent
  in a package, but just as a raw string ending with \r\n. The RFC also
  says that this string can be prefixed with other lines ending in \r\n. 

  The RFC also doesn't prohibit the server from immediately sending a
  KEX_INIT message after the identification string. So, if we read too
  much from the socket, we will end up reading part of the KEX_INIT
  message. So, this requires some care.
 ***********************************************************************/



static int tssh_connect(tssh_t *T)
{
  int R;
  int i;
  struct sockaddr_in addr;
  int fd;
  struct addrinfo hints, *list, *iter;
  char *service;
  char servicebuf[20];
  char rhost[255]; char rservice[255];

  memset(&hints,0,sizeof(hints));
  hints.ai_family= AF_UNSPEC;
  hints.ai_socktype= SOCK_STREAM;
  hints.ai_protocol= 0;
  hints.ai_flags= AI_ADDRCONFIG | AI_V4MAPPED | AI_ALL | AI_NUMERICSERV;
  if (T->port>0) { service= servicebuf; sprintf(servicebuf,"%d",T->port); }
            else { service= NULL; }

  R= getaddrinfo(T->host, service, &hints, &list);
  switch(R)
  {
  case 0:
    break;
  case EAI_SYSTEM:
    tssh_log(T, "getaddrinfo(): %s\n", strerror(errno));
    return 1;
  default:
    tssh_log(T, "getaddrinfo(): %s\n", gai_strerror(errno));
    return 1;
  }

  fd= -1;
  for(i=0,iter=list;iter;iter=iter->ai_next,i++)
  {
    if (getnameinfo(iter->ai_addr, iter->ai_addrlen, 
                    rhost, sizeof(rhost), rservice, sizeof(rservice),
                    NI_NUMERICHOST | NI_NUMERICSERV))
      tssh_log(T, "trying address #%d\n", i);
    else
      tssh_log(T, "trying %s/%s\n", rhost, rservice);

    fd= socket(iter->ai_family, iter->ai_socktype, iter->ai_protocol);
    if (fd<0) 
    {
      tssh_log(T, "socket(): %s\n", strerror(errno));
      continue;
    }
 
    if (connect(fd, iter->ai_addr, iter->ai_addrlen))
    {
      tssh_log(T, "connect(): %s\n", strerror(errno));
      close(fd);
      fd= -1;
      continue;
    }
    break;
  }
  freeaddrinfo(list);
  if (fd<0) return 1;
  T->fd= fd;
  return 0;
}


static int tssh_send(tssh_t *T, unsigned char *buf, int len)
{
  int  R;
  while(len)
  {
    R= send(T->fd, buf, len, MSG_NOSIGNAL);
    if (R<0)
    {
      if (errno==EINTR) continue;
      tssh_log(T, "send(): %s\n", strerror(errno));
      return 1;
    }
    buf+= R;
    len-= R;
  }
  return 0;
}

static int tssh_recv(tssh_t *T, unsigned char *buf, int len)
{
  int R;
  do R= recv(T->fd, buf, len, 0);
  while(R<0 && errno==EINTR);
  if (R==0) return -1;
  return R;
}
