
/**
  The integrity check also makes use of hash functions, but a specific
  algorithm is used for creating the hash. It's not just a hash of
  the message contents. 

  This algorithm is defined in RFC-2104.
 **/

static void tssh_hmac_begin(tssh_hmac_t *M,uint32_t seqno)
{
  uint8_t buf[4];
  tssh_put_uint32(buf, seqno);
  switch(M->type)
  {
  case TSSH_SHA1:
    SHA1Init(&M->C.sha1);
    SHA1Update(&M->C.sha1, M->key_ipad, 64);
    SHA1Update(&M->C.sha1, buf, 4);
    break;
  }
}

static void tssh_hmac_update(tssh_hmac_t *M,uint8_t *data, uint32_t len)
{
  switch(M->type)
  {
  case TSSH_SHA1:
    SHA1Update(&M->C.sha1, data, len);
    break;
  }
}

static void tssh_hmac_finish(tssh_hmac_t* M)
{
  // WARNING: when you add a new algorithm with a bigger
  // digest length, grow this variable accordingly.
  uint8_t prelim[20];
  switch(M->type)
  {
  case TSSH_SHA1:
    SHA1Final(prelim, &M->C.sha1);
    SHA1Init(&M->C.sha1);
    SHA1Update(&M->C.sha1, M->key_opad, 64);
    SHA1Update(&M->C.sha1, prelim, 20);
    SHA1Final(M->data, &M->C.sha1);
    break;
  }
}

static void tssh_hmac_init(tssh_hmac_t *M,int type,uint8_t *key)
{
  int i;
  M->type= type;
  switch(type)
  {
  case TSSH_SHA1:
    M->keylen= 20;
    M->key_ipad= calloc(1,64);
    memcpy(M->key_ipad, key, 20);
    for(i=0;i<64;i++) M->key_ipad[i]^= 0x36;

    M->key_opad= calloc(1,64);
    memcpy(M->key_opad, key, 20);
    for(i=0;i<64;i++) M->key_opad[i]^= 0x5c;

    M->len= 20;
    M->data= malloc(20);
    break;
  }
}

static struct tssh_hmac_def tssh_hmac_defs[]=
{
  { "hmac-sha1", 20, TSSH_SHA1 },
  { NULL }
};

static struct tssh_hmac_def *tssh_find_hmac_def(char *name)
{
  struct tssh_hmac_def *D;
  for(D=tssh_hmac_defs;D->name;D++)
   if (!strcmp(D->name,name)) return D;
  return NULL;
}


static struct tssh_hmac_def *tssh_nego_hmac_def(tssh_obj_t *serv_o)
{
  tssh_obj_t server_list, server;
  struct tssh_hmac_def *D;

  for(D=tssh_hmac_defs;D->name;D++)
  {
    server_list= *serv_o;
    tssh_namelist_skiplen(&server_list);
    while(tssh_namelist_get(&server_list, &server))
    {
      if (server.len!=strlen(D->name)) continue;
      if (memcmp(server.data,D->name,server.len)) continue;
      return D;
    }
  }
  return NULL;
}

static char *tssh_hmac_names()
{
  static char *R;
  int n;
  size_t L;
  buffer_t *B;
  struct tssh_hmac_def *D;

  if (R) return R;
  B= buffer_new(1024);
  for(D=tssh_hmac_defs,n=0;D->name;D++,n++)
  {
    if (n) buffer_write(B, ",", 1);
    buffer_write(B, D->name, strlen(D->name));
  }
  buffer_write(B, "", 1);
  buffer_free(B, (void**) &R, &L);
  return R;
}


