
/***********************************************************************
 *                                                                     *
 *                            Writer Thread                            *
 *                                                                     *
 ***********************************************************************
  Outbound messages don't require any type of flow control except
  for the SSH_CHANNEL_DATA message. This message requires it because
  of the reason explained below.

  An SSH connection is really a multitude of connections. Thru just one
  socket you have a connection to both the SSH server(S) and the program 
  (A) handling your channel data. When you send some channel data to S,
  it has to deliver the data to A. A may not be ready to recieve it
  immediately. There are three options for S:
    1- stall until A is ready
    2- buffer the data
    3- drop the data
  Options 1 and 3 are unacceptable. In case of a stall, S would not be
  able to process any incoming control messages. In order to limit the
  size of the channel receive buffer on S, we have to do flow control.
  For this reason, a window_size is used, effectively telling us the
  remaining space in the channel receive buffer on S.

  When the remote party (A) has stalled, we have to buffer the data and
  eventually stall the user thread. This is done as follows:

  When we need to send channel data, we try to put it in the channel's
  buffer. If the buffer already has too much data in it, we wait on a
  condition variable. After that, we signal the main thread with
  TSSH_CMD_CHANNEL_WRITE. This is done by the tssh_channel_write_buffer
  function.

  Main thread receives the TSSH_CMD_CHANNEL_WRITE command. It checks 
  whether the channel has enough window space. If there is no space, 
  the command is ignored, thus postponing the write operation. If there
  is space, the command is forwarded to the writer thread.

  The writer thread encodes the packet and sends it off. After that,
  it signals the main thread with TSSH_CMD_CHANNEL_WRITE_OK. This tells
  the main thread to check again, whether there is any more data to send.
  If so, the process continues with another TSSH_CMD_CHANNEL_WRITE to
  the writer thread.

  The advantage of this method is that only the main thread accesses
  the variable channel->window_size_c2s. This ensures that incoming
  SSH_WINDOW_ADJUST messages are processed properly.

  When the main thread finishes sending one tssh_write_buffer_t, it
  sends a CHANNEL_WRITE_OK event to the user. Each buffer corresponds
  to a tssh_channel_write() call, so this tells the user that the
  last such call was successful. This event tells the user that another
  channel_write() would probably not stall at this time.
 ***********************************************************************/

static void tssh_opak_bytes(tssh_opak_t *T, uint8_t *data, uint32_t len)
{
  buffer_write(T->buffer, data, len);
  T->len+= len;
}

static void tssh_opak_uint32(tssh_opak_t *T, uint32_t V)
{
  unsigned char buf[4];
  tssh_put_uint32(buf, V);
  tssh_opak_bytes(T, buf, 4);
}

static void tssh_opak_string(tssh_opak_t *T,char *data)
{
  uint32_t L;
  L= strlen(data);
  tssh_opak_uint32(T, L);
  tssh_opak_bytes(T, data, L);
}

static void tssh_opak_namelist(tssh_opak_t *T, char *data)
{
  tssh_opak_string(T,data);
}

static void tssh_opak_byte(tssh_opak_t *T, unsigned char V)
{
  buffer_write(T->buffer, &V, 1);
  T->len+= 1;
}

static void tssh_opak_bool(tssh_opak_t *T, int v)
{
  tssh_opak_byte(T,v);
}

static tssh_opak_t* tssh_opak_begin()
{
  tssh_opak_t *T;
  T= malloc(sizeof(*T));
  T->buffer= buffer_new(1024);
  T->len= 0;
  return T;
}

static int tssh_encode_send
   (tssh_t *T, tssh_encoder_t *encoder, uint8_t *buffer, int len)
{
  // tssh_debug_hex(buffer, len, "output");
  if (encoder) 
  {
    tssh_hmac_update(&encoder->hmac, buffer, len);
    tssh_encrypt(&encoder->crypto, buffer,len);
  }
  return tssh_send(T, buffer, len);
}


static void tssh_writer_thread(tssh_t *T)
{
  void *ap1, *ap2, *ap3; uint32_t ai1; int cmd;
  tssh_encoder_t *encoder;
  tssh_opak_t *P;
  uint32_t total_len, i;
  uint8_t  padding_len, extra_padding_len;
  uint32_t  payload_len, buflen, bufsiz, cp;
  uint32_t packet_no;
  uint8_t *write_buffer;
  uint32_t write_buffer_size;
  int oQ;

  // these are for the channel writer
  uint8_t *hdr;
  uint8_t *data;
  uint32_t packet_len, data_len;
  tssh_channel_t *ch;
  tssh_buffer_t *buffer_to_free;
  uint32_t amount;

  packet_no= 0xFFffFFff;
  encoder= NULL;
  write_buffer_size= 1024;
  write_buffer= malloc(write_buffer_size);
  oQ= 8;

again:
  buffer_to_free= NULL;
  tssh_get_thread_msg(&T->writer_queue, &cmd, &ap1, &ap2, &ai1, &ap3);
  switch(cmd)
  {
  case TSSH_CMD_QUIT: 
    if (encoder) free(encoder);
    free(write_buffer);
    return ;
  case TSSH_CMD_SETENCODER: 
    if (encoder) free(encoder);
    encoder= ap1;
    oQ= encoder->cryptodef->block_len;
    goto again;
  case TSSH_CMD_WRITE:
    break;
  case TSSH_CMD_CHANNEL_WRITE:
    ch= ap1;
    data= ap2;
    data_len= ai1;
    buffer_to_free= ap3;
    goto write_channel;
  default:
    goto again;
  }
  P= ap1;

  packet_no++;
  payload_len= P->len;
  total_len= 5 + payload_len +4 ;  // 5 bytes header, minimum 4 bytes padding
  extra_padding_len=  oQ - total_len%oQ;
  padding_len= 4 + extra_padding_len;
  total_len += extra_padding_len;

  tssh_put_uint32(write_buffer, total_len - 4);
  tssh_put_uint8(write_buffer+4, padding_len);

  buflen= 5;
  bufsiz= write_buffer_size - write_buffer_size%oQ;

  if (encoder) tssh_hmac_begin(&encoder->hmac, packet_no);

  while(payload_len)
  {
    cp= bufsiz - buflen;
    if (cp>payload_len) cp= payload_len;

    buffer_read(P->buffer, write_buffer+buflen, cp);
    payload_len-= cp;
    buflen += cp;

    // we have run out of payload
    if (buflen%oQ) break; 
    
    if (tssh_encode_send(T, encoder, write_buffer, buflen)) goto fail;
    buflen= 0;
  }

  // careful: we might have a whole block of padding
  // after we finish off this one.
  for(;buflen%oQ;buflen++) { write_buffer[buflen]= rand();
                                padding_len--; }

  if (tssh_encode_send(T, encoder, write_buffer, buflen)) goto fail;
  
  if (padding_len) // this should be the same as oQ..
  {
    for(buflen=0;buflen<padding_len;buflen++)
      write_buffer[buflen]= rand();

    if (tssh_encode_send(T, encoder, write_buffer, buflen)) goto fail;
  }

  if (encoder)
  {
    tssh_hmac_finish(&encoder->hmac);
    // tssh_debug_hex(encoder->hmac.data, encoder->hmac.len, "hmac-out");
    if (tssh_send(T, encoder->hmac.data,encoder->hmac.len)) goto fail;
  }
  buffer_free(P->buffer, NULL, NULL);
  free(P);
  goto again;
fail:
  if (buffer_to_free) free(buffer_to_free);
  // the only possible failure case comes from the send() call.
  // we send an error command to the main thread.
  // FIXME: actually do it
  return ;

write_channel:

  packet_no++;
  if (encoder) tssh_hmac_begin(&encoder->hmac, packet_no);

  // data always has space before in which the header will fit.
  hdr= data - TSSH_CHANNEL_DATA_HDRSIZE;
  
  packet_len= tssh_outbound_channel_data_packet_size(oQ,data_len,&padding_len);
  tssh_put_uint32(hdr, packet_len-4);
  tssh_put_uint8(hdr+4, padding_len );
  tssh_put_uint8(hdr+5, SSH_MSG_CHANNEL_DATA);
  tssh_put_uint32(hdr+6, ch->remote_number);
  tssh_put_uint32(hdr+10, data_len);

  data= hdr;
  total_len= data_len + TSSH_CHANNEL_DATA_HDRSIZE;
  amount= total_len - total_len%oQ;

  if (tssh_encode_send(T, encoder, data, amount)) goto fail;
  data+= amount;
  total_len -= amount;
  if (total_len) memcpy(write_buffer, data, total_len);
  for(i=0;i<padding_len;i++) write_buffer[total_len+i]= rand();
  if (tssh_encode_send(T, encoder, write_buffer, total_len+padding_len))
     goto fail;

  if (encoder)
  {
    tssh_hmac_finish(&encoder->hmac);
    if (tssh_send(T, encoder->hmac.data,encoder->hmac.len)) goto fail;
  }
  if (buffer_to_free) free(buffer_to_free);
  tssh_put_thread_msg(&T->main_queue, 0, TSSH_CMD_CHANNEL_WRITE_OK,
                        ch, NULL, data_len, NULL);
  goto again;
}

static void *tssh_writer_thread_entry(void *arg)
{
  tssh_writer_thread(arg);
  return NULL;
}

static void tssh_send_opak(tssh_t *T, tssh_opak_t *P)
{
  tssh_put_thread_msg(&T->writer_queue, 0, TSSH_CMD_WRITE, P, NULL, 0, 0);
}

static tssh_obj_t *tssh_opak_to_obj(tssh_opak_t *P)
{
  uint8_t *data;
  size_t len;
  buffer_collect(P->buffer, (void**) &data, &len);
  return tssh_obj_new(data,len);
}

static void tssh_opak_obj(tssh_opak_t *P, tssh_obj_t *O)
{
  tssh_opak_bytes(P, O->data, O->len);
}

static void tssh_wb_put_uint8(tssh_write_buffer_t *wb, uint8_t V)
{
  tssh_put_uint8(wb->data+wb->len, V);
  wb->len+= 1;
}

static void tssh_wb_put_uint16(tssh_write_buffer_t *wb, uint16_t V)
{
  tssh_put_uint16(wb->data+wb->len, V);
  wb->len+= 2;
}


static void tssh_wb_put_uint32(tssh_write_buffer_t *wb, uint32_t V)
{
  tssh_put_uint32(wb->data+wb->len, V);
  wb->len+= 4;
}

static void tssh_wb_put_uint64(tssh_write_buffer_t *wb, uint64_t V)
{
  tssh_put_uint64(wb->data+wb->len, V);
  wb->len+= 8;
}

static void tssh_wb_put_string(tssh_write_buffer_t *wb, char *str)
{
  uint32_t L;
  L= strlen(str);
  tssh_wb_put_uint32(wb, L);
  memcpy(wb->data+wb->len, str, L);
  wb->len+= L;
}

static void tssh_wb_put_obj(tssh_write_buffer_t *wb, tssh_obj_t *obj)
{
  memcpy(wb->data+wb->len, obj->data, obj->len);
  wb->len += obj->len;
}
