

static void tssh_rfcstr_skiplen(tssh_obj_t* ob)
{
  ob->data+= 4;
  ob->len-= 4;
}

static void tssh_rfcstr_get(tssh_obj_t *src, tssh_obj_t *dst)
{
  uint32_t L;
  L= tssh_get_uint32(src->data);

  dst->data= src->data;
  dst->len= L+4;

  src->data+= L+4;
  src->len-= L+4;
}

static void tssh_rfcmpint_get(tssh_obj_t *src, tssh_obj_t *dst)
{
  tssh_rfcstr_get(src, dst);
}

static int tssh_strobj_match(tssh_obj_t *src, char *v)
{
  int L;
  L= strlen(v);
  if (src->len!=L+4) return 0;
  if (memcmp(src->data+4, v, L)) return 0;
  return 1;
}

/***
  host key looks like:

   string[ string[ssh-rsa] || mpint[e] || mpint[n] ]

  the signature looks like:

   string[ string[ssh-rsa] || mpint[s] ]

  As I understand it, I should have:

    (s^e)%n == {H}

  {H} is the encoded form of H.
  {H}= rsa_padding || hash_prefix || hash(H)

  here, hash_prefix is an ASN.1 object descriptor identifying
  the hashed value..
  ***/

static tssh_mpint_t* tssh_decrypt_signature(tssh_t *T)
{
  tssh_obj_t signature, host_key, sigval, tag;
  tssh_obj_t sval, nval, eval;
  tssh_mpint_t *e,*s,*n,*he;
  tssh_numbers_t *table;

  signature= *T->kex->signature;
  host_key= *T->kex->host_key;

  tssh_rfcstr_skiplen(&signature);
  tssh_rfcstr_get(&signature, &tag);
  if (!tssh_strobj_match(&tag, "ssh-rsa")) return NULL;
  tssh_rfcmpint_get(&signature, &sval);

  tssh_rfcstr_skiplen(&host_key);
  tssh_rfcstr_get(&host_key, &tag);
  if (!tssh_strobj_match(&tag, "ssh-rsa")) return NULL;
  tssh_rfcmpint_get(&host_key, &eval);
  tssh_rfcmpint_get(&host_key, &nval);

  e= tssh_mpint_from_rfc_mpint(eval.data);
  n= tssh_mpint_from_rfc_mpint(nval.data);
  s= tssh_mpint_from_rfc_mpint(sval.data);

  //tssh_mpint_print("VFY e= ", e);
  //tssh_mpint_print("VFY n= ", n);
  //tssh_mpint_print("VFY s= ", s);

  if (!tssh_mpint_greater(n,s)) 
  {
    he= NULL;
    goto done;
  }

  table= tssh_multiples_of_modulus(n->data,n->len);
  he= tssh_expmod_by_table(s, e, table);
  tssh_numbers_free(table);
  // tssh_mpint_print("VFY he=", he);
done:
  tssh_mpint_free(e);
  tssh_mpint_free(s);
  tssh_mpint_free(n);
  return he;
}

static int tssh_verify_signature(tssh_t *T)
{
  tssh_mpint_t *he;  // H encoded
  static const uint8_t pfx_sha1[] = {
    0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02,
    0x1a, 0x05, 0x00, 0x04, 0x14 };         
  const uint8_t *pfx; uint32_t pfx_len;
  uint8_t *data; uint32_t data_len;
  uint8_t *start;
  tssh_hash_t *hash;
  char *errstr;

  he= tssh_decrypt_signature(T);
  if (!he) 
  {
    tssh_log(T,"bad signature in message\n");
    return 1;
  }

  tssh_mpint_to_rfc_mpint(he, &data_len, &data);
  tssh_mpint_free(he);

  pfx= pfx_sha1; pfx_len= sizeof(pfx_sha1);
  hash= tssh_hash_alloc(TSSH_SHA1);

  errstr= "result of decryption too short in rsa-verify\n";
  if (data_len<hash->len+pfx_len) goto fail;

  errstr= "signature doesn't match\n";
  start= data + data_len - pfx_len - hash->len; 
  if (memcmp(start, pfx, pfx_len))  goto fail;

  start+= pfx_len;
  tssh_hash_begin(hash);
  tssh_hash_obj(hash, T->kex->H);
  tssh_hash_finish(hash);
  if (memcmp(start, hash->data, hash->len)) goto fail;

  tssh_hash_dealloc(hash);
  free(data);
  return 0;
fail:
  tssh_hash_dealloc(hash);
  free(data);
  tssh_log(T, errstr);
  return 1;
}
