
static tssh_ftop_t *tssh_ftop_new(int code)
{
  tssh_ftop_t *op;
  op= calloc(1,sizeof(*op));
  op->cmd= code;
  return op;
}

static void tssh_ftop_free(tssh_ftop_t *op)
{
  if (op->fn) free(op->fn);
  if (op->oldfn) free(op->oldfn);
  if (op->data) free(op->data);
  if (op->localfn) free(op->localfn);
  free(op);
}

static void tssh_ftp_report_opfail(tssh_t *T, tssh_ftop_t *op)
{
  tssh_event_t *E;
  E= tssh_event_new(TSSH_EVENT_FTP_FAILURE);
  E->ftp_op= op;
  tssh_report(T, E);
}

static void tssh_ftp_report_all_opfail(tssh_t *T)
{
  tssh_ftop_t *op;
  if (T->ftop_current) tssh_ftp_report_opfail(T, T->ftop_current);
  for(op=T->ftop_first;op;op=op->next)
   tssh_ftp_report_opfail(T, T->ftop_current);
}

static int tssh_ftp_report_success_simple(tssh_t *T)
{
  tssh_event_t *E;
  E= tssh_event_new(TSSH_EVENT_FTP_SUCCESS);
  E->ftp_op= T->ftop_current;
  T->ftop_current= NULL;
  tssh_report(T, E);
  return TSSH_OK;
}

static void tssh_ftp_channel_destroy(tssh_t *T)
{
  if (!T->ftp_channel) return ;
  
  tssh_send_channel_close(T, T->ftp_channel);
  tssh_ftp_mqdestroy(&T->ftp_channel->ftp_mq);
  free(T->ftp_channel);
  T->ftp_channel= NULL;
}

// This function never returns. 
static void tssh_ftp_die(tssh_t *T,int ev)
{
  coth_t *ctx;
 
  tssh_ftp_channel_destroy(T);
  tssh_ftp_report_all_opfail(T);
  tssh_report_simple(T, ev);

  ctx= T->ftp_ctx;
  T->ftp_ctx= NULL;
  tssh_put_thread_msg(&T->main_queue, 0, 
                      TSSH_CMD_DEALLOC_COTH, ctx, NULL,0,NULL);
  coth_switch(ctx, T->main_ctx);
}
//
// This function is called the when creation of the ftp channel is confirmed
// by the server.
//
static void tssh_ftp_channel_open_success(tssh_t *T)
{
  tssh_log(T, "ftp channel created\n");
  tssh_ftp_thread_msg(T, TSSH_CMD_FTP_CHANNEL_CREATE_OK, NULL);
}

static void tssh_ftp_channel_open_failure(tssh_t *T)
{
  tssh_log(T, "ftp channel creation failed\n");
  tssh_ftp_thread_msg(T, TSSH_CMD_FTP_CHANNEL_CREATE_FAIL, NULL);
}

//
// this function gets called when a channel request finishes successfully
//
static void tssh_ftp_channel_success(tssh_t *T)
{
  tssh_log(T, "ftp subsystem started\n");
  tssh_ftp_thread_msg(T, TSSH_CMD_FTP_CHANNEL_REQUEST_OK, NULL);
}

//
// this function gets called when a channel request fails
//
static void tssh_ftp_channel_failure(tssh_t *T)
{
  tssh_log(T, "ftp channel: subsystem initialization failed.\n");
  tssh_ftp_thread_msg(T, TSSH_CMD_FTP_CHANNEL_REQUEST_FAIL, NULL);
}

//
// this is called when the remote end sends a channel_close message.
// on entry, the channel is already closed (msg sent, data freed).
// all we have to do is to report the fact to the ftp thread. it will
// then close the necessary file descriptors and wait to be expunged.
//
static void tssh_ftp_channel_closed(tssh_t *T)
{
  tssh_log(T, "ftp channel closed\n");
  tssh_ftp_mqdestroy(&T->ftp_channel->ftp_mq);
  free(T->ftp_channel);
  T->ftp_channel= NULL;
  tssh_ftp_thread_msg(T, TSSH_CMD_FTP_DISCONNECT, NULL);
}

static int tssh_ftp_wait
   (tssh_t *T, int *Rcmd, void **Rdata, int nrep, ... )
{
  va_list ap;
  int i;
  int cmd; void *data;

restart:
  if (!T->ftp_ctx) return 1;
  if (coth_msg_get(T->ftp_ctx, &cmd, &data)) 
  {
    coth_switch(T->ftp_ctx, T->main_ctx);
    goto restart;
  }

  if (cmd==TSSH_CMD_FTP_DISCONNECT) return 1;

  if (!nrep) goto fail;
  va_start(ap, nrep);
  for(i=0;i<nrep;i++)
  {
    if (va_arg(ap,int)!=cmd) continue;
    if (Rcmd) *Rcmd= cmd;
    if (Rdata) *Rdata= data;
    va_end(ap);
    return 0;
  }
  va_end(ap);
fail:
  //
  // the only kind of data we can receive here is a tssh_buffer_t, 
  // for the FTP_DATA command. that can be directly free()d.
  //
  if (data) free(data);
  return 1;
}

static int tssh_ftp_wait_data
   (tssh_t *T, int *Rcmd, tssh_buffer_t **Rdata, int nrep, ...)
{
  int i;
  va_list ap;
  void *data;
  tssh_buffer_t *B;
  int cmd;
  uint8_t cc;

  data= NULL;
  if (tssh_ftp_wait(T,  &cmd, &data, 1, TSSH_CMD_FTP_DATA)) goto fail;

  B= data;
  if (!B) return 1;
  if (tssh_xparse_sftpcmd("f-cmd", B, &cc))  goto fail;

  va_start(ap, nrep);
  for(i=0;i<nrep;i++)
  {
    if (va_arg(ap, int)!=cc) continue;
    if (Rcmd) *Rcmd= cc;
    if (Rdata) *Rdata= B;
    va_end(ap);
    return 0;
  }

  va_end(ap);
fail:
  if (data) free(data);
  return 1;
}

static void tssh_ftp_start(tssh_t *T)
{
  if (T->ftp_ctx) return ;
  T->ftp_ctx= coth_new(tssh_ftp_main, 40960);
  coth_arg_p(T->ftp_ctx, 1, T);
  coth_switch(T->main_ctx, T->ftp_ctx);
}

static int tssh_ftp_channel_data(tssh_t *T, tssh_buffer_t *pb)
{
  tssh_buffer_t *B;
  tssh_channel_t *ch;
  ch= T->ftp_channel;
  if (!ch || !T->ftp_ctx)
  {
    free(pb);
    return 0;
  }
  tssh_ftp_mqput(&ch->ftp_mq, pb);
  while((B=tssh_ftp_mqget(&ch->ftp_mq)))
    coth_msg_put(T->ftp_ctx, TSSH_CMD_FTP_DATA, B);
  coth_switch(T->main_ctx, T->ftp_ctx);
  return 0;
}
