
// upon entry, we don't even have the ssh channel yet.
static void tssh_ftp_do_connect(tssh_t *T)
{
  tssh_opak_t *O;
  int cmd;
  tssh_buffer_t *response;
  uint32_t version;

  T->ftp_channel= tssh_channel_create(T, TSSH_CHANNEL_FTP);
  O= tssh_channel_open_opak("session", T->ftp_channel);
  tssh_send_opak(T,O);

  if (tssh_ftp_wait(T, &cmd, NULL, 2,
                      TSSH_CMD_FTP_CHANNEL_CREATE_OK, 
                      TSSH_CMD_FTP_CHANNEL_CREATE_FAIL) ||
      cmd==TSSH_CMD_FTP_CHANNEL_CREATE_FAIL)
    goto fail;

  O= tssh_channel_request(T->ftp_channel, 1, "subsystem");
  tssh_opak_string(O, "sftp");
  tssh_send_opak(T, O);  

  if (tssh_ftp_wait(T, &cmd, (void**) NULL, 2,
                       TSSH_CMD_FTP_CHANNEL_REQUEST_OK,
                       TSSH_CMD_FTP_CHANNEL_REQUEST_FAIL) ||
     cmd==TSSH_CMD_FTP_CHANNEL_REQUEST_FAIL)
    goto fail;

  if (tssh_ftp_cmd(T, "Cd", SSH_FXP_INIT, 3))
     goto fail;
  if (tssh_ftp_wait_data(T, &cmd, &response, 1,
                         SSH_FXP_VERSION)) goto fail;

  if (tssh_xparse_uint32("f-version", response, &version))
  {
    free(response);
    tssh_log(T, "bad ftp-version message\n");
    goto fail;
  }

  if (version!=3)
  {
    free(response);
    tssh_log(T, "incompatible sftp version %d reported by server.\n",version);
    goto fail;
  }

  tssh_log(T, "SFTP Server version %d\n", version);   

#ifdef DEBUG
  while(response->len)
  {
    tssh_obj_t name, data;
    tssh_parse_extpair(response, &name, &data);
    printf("ext["); fwrite(name.data+4, 1, name.len-4, stdout);
    printf(")\n");
  }
#endif                        
  free(response);
  tssh_report_simple(T, TSSH_EVENT_FTP_CONNECT_SUCCESS); 
  return ;
fail:
  tssh_ftp_die(T, TSSH_EVENT_FTP_CONNECT_FAILURE);
}


static tssh_obj_t *tssh_ftp_get_handle(tssh_buffer_t *res)
{
  tssh_obj_t handle;
  if (tssh_xparse_string("f-handle", res, &handle)) return NULL;
  return tssh_objdup(&handle);
}

static int tssh_ftp_report_status(tssh_t *T, tssh_buffer_t *res)
{
  uint32_t st_code;
  tssh_obj_t st_msg;
  uint32_t reqid;

  if (tssh_xparse_uint32("f-reqid", res, &reqid) || 
      tssh_xparse_uint32("f-code", res, &st_code) ||
      tssh_xparse_string("f-msg", res, &st_msg))
  {
    tssh_log(T,"ftp: bad status message\n");
    return SSH_FX_BAD_MESSAGE;
  }
  if (st_code!=SSH_FX_OK && st_code!=SSH_FX_EOF)
    tssh_log(T,"ftp: %S\n", &st_msg);
  return st_code;
}

static int tssh_ftp_close_handle(tssh_t *T, tssh_obj_t *handle)
{
  int cmd;
  tssh_buffer_t *res;
  int retval;

  if (!handle) return 0;

  res= NULL;
  retval= TSSH_OK;
  if (tssh_ftp_cmd(T, "CRo", SSH_FXP_CLOSE, handle) || 
      tssh_ftp_wait_data(T, &cmd, &res, 1, SSH_FXP_STATUS)) 
          retval= TSSH_FAIL_HARD;
  else
     if (tssh_ftp_report_status(T, res)!=SSH_FX_OK) retval= TSSH_FAIL;

  if (res) free(res);
  tssh_obj_free_deep(handle);
  return retval;
}

static int tssh_ftp_open_file
  (tssh_t *T, tssh_ftop_t* op, int rW, uint32_t pflags, uint32_t perm,
   tssh_obj_t **Rhandle)
{
  int retval;
  int cmd;
  tssh_buffer_t *res;
  tssh_obj_t *handle;
  uint32_t reqid;

  handle= NULL;
  retval= TSSH_FAIL;
  res= NULL;

  if (rW)
  {
   if (tssh_ftp_cmd(T, "CRsddd", SSH_FXP_OPEN, op->fn, 
                           pflags, 
                            SSH_FILEXFER_ATTR_PERMISSIONS, perm)) 
     goto fail_hard;
  }
  else
  {
   if (tssh_ftp_cmd(T, "CRsdd", SSH_FXP_OPEN, op->fn, 
                           pflags, 0))
     goto fail_hard;
  }

  if (tssh_ftp_wait_data(T, &cmd, &res, 2, SSH_FXP_HANDLE, SSH_FXP_STATUS))
    goto fail_hard;

  if (cmd==SSH_FXP_STATUS)
  {
    tssh_ftp_report_status(T, res);
    goto fail;
  }

  if (tssh_xparse_uint32("f-reqid", res, &reqid) || 
      !(handle=tssh_ftp_get_handle(res)))
    goto fail_hard;

  free(res);
  *Rhandle= handle;
  return TSSH_OK;

fail_hard:
  retval= TSSH_FAIL_HARD;
fail:
  if (res) free(res);
  if (handle) tssh_obj_free_deep(handle);
  return retval;
}

static int tssh_ftp_read_file
  (tssh_t *T, tssh_obj_t *handle, uint64_t *offset, 
   uint32_t size, tssh_buffer_t **Rdata)
{
  tssh_buffer_t *res;
  int cmd;
  uint32_t reqid;
  uint32_t len;

  if (tssh_ftp_cmd(T, "CRoqd", SSH_FXP_READ, handle, *offset, size))
    return TSSH_FAIL_HARD;
  if (tssh_ftp_wait_data(T, &cmd, &res, 2, SSH_FXP_DATA, SSH_FXP_STATUS))
    return TSSH_FAIL_HARD;

  if (cmd==SSH_FXP_STATUS)
  {
    if (tssh_ftp_report_status(T, res)==SSH_FX_EOF)
    {
      free(res);
      return TSSH_EOF;
    }
    free(res);
    return TSSH_FAIL;
  }

  if (tssh_xparse_uint32("f-reqid", res, &reqid)) goto bad_msg;
  if (tssh_xparse_uint32("f-len", res, &len)) goto bad_msg;
  if (len!=res->len) goto bad_msg;
  *Rdata= res;
  *offset += len;
  return TSSH_OK;

bad_msg:
  free(res);
  tssh_log(T, "bad FXP_DATA message\n");
  return TSSH_FAIL_HARD;
}

static int tssh_ftp_write_file
  (tssh_t *T, tssh_obj_t *handle, uint64_t *offset, 
   uint8_t *data, uint32_t len)
{
  tssh_buffer_t *res;
  int cmd;
  int retval;

  retval= TSSH_OK;
  if (tssh_ftp_cmd(T, "CRoqdB", SSH_FXP_WRITE, handle, *offset, len, len, data))
    return TSSH_FAIL_HARD;
  if (tssh_ftp_wait_data(T, &cmd, &res, 1, SSH_FXP_STATUS))
    return TSSH_FAIL_HARD;

  if (tssh_ftp_report_status(T, res)!=SSH_FX_OK) retval= TSSH_FAIL;
  *offset += len;
  free(res);
  return retval;
}

static int tssh_ftp_do_list(tssh_t *T)
{
  int cmd;
  int retval;
  tssh_buffer_t *res;
  buffer_t *buffer;
  tssh_obj_t *handle, obj;
  uint32_t count, i, mcount;
  uint32_t reqid;
  tssh_event_t *E;
  tssh_fileinfo_t finfo;
  tssh_obj_t name, lname;  
  tssh_fileinfo_t *finfos;

  retval= TSSH_FAIL;
  res= NULL;
  buffer= NULL;

  if (tssh_ftp_cmd(T, "CRs", SSH_FXP_OPENDIR, T->ftop_current->fn))
    goto fail_hard;

  if (tssh_ftp_wait_data(T, &cmd, &res, 2, 
      SSH_FXP_HANDLE, SSH_FXP_STATUS)) goto fail_hard;

  if (cmd!=SSH_FXP_HANDLE) 
  {
    tssh_ftp_report_status(T,res);
    goto fail;
  }

  if (tssh_xparse_uint32("f-reqid", res, &reqid)) goto bad_response;
  if (!(handle=tssh_ftp_get_handle(res))) goto bad_response;

  free(res); res= NULL;

  buffer= buffer_new(sizeof(tssh_fileinfo_t)*50);
  count= 0;
  
read_more:
  if (tssh_ftp_cmd(T, "CRo", SSH_FXP_READDIR, handle)) goto fail_hard;

  if (tssh_ftp_wait_data(T, &cmd, &res, 2, 
      SSH_FXP_NAME, SSH_FXP_STATUS)) goto fail_hard;

  switch(cmd)
  {
  case SSH_FXP_NAME:
    if (tssh_xparse_uint32("f-reqid", res, &reqid)) goto bad_response;
    if (tssh_xparse_uint32("f-count", res, &mcount)) goto bad_response;
    for(i=0;i<mcount;i++)
    {
       if (tssh_xparse_string("  name", res, &name)) goto bad_response;
       if (tssh_xparse_string("  l-name", res, &lname)) goto bad_response;  
       if (tssh_ftp_parse_file_attr(T, res, &finfo)) goto bad_response;
       finfo.name= tssh_xstr_to_str(&name);
#ifndef TSSH_NODEBUG
       fwrite(lname.data+4, 1, lname.len-4, stdout);
       printf("\n");
#endif
       buffer_write(buffer, &finfo, sizeof(finfo));
       count++;
    }
    free(res); res= NULL;
    goto read_more;
  case SSH_FXP_STATUS:
    if (tssh_ftp_report_status(T, res)!=SSH_FX_EOF) goto fail;
    break;
  }
  free(res); res= NULL;

  switch(tssh_ftp_close_handle(T, handle)) 
  {
  case TSSH_FAIL: goto fail;
  case TSSH_FAIL_HARD: goto fail_hard;
  }
  
  buffer_free(buffer, (void**) &finfos, NULL);
  E= tssh_event_new(TSSH_EVENT_FTP_SUCCESS);
  E->n_fileinfo= count;
  E->fileinfo= finfos;
  E->ftp_op= T->ftop_current;
  T->ftop_current= NULL;
  tssh_report(T, E);
  return 0;

bad_response:
  tssh_log(T, "ftp-list: bad response\n");
fail_hard:
  retval= TSSH_FAIL_HARD;
fail:
  if (res) free(res);
  if (tssh_ftp_close_handle(T, handle)==TSSH_FAIL_HARD) 
     retval= TSSH_FAIL_HARD;
  if (buffer)
  {
    buffer_free(buffer, (void**) finfos, NULL);
    tssh_free_finfos(finfos, count);
  }
  return retval; 
}

static int tssh_ftp_do_downloadmem(tssh_t *T)
{
  int retval;
  tssh_obj_t *handle;
  int R;
  tssh_buffer_t *res;
  buffer_t *buffer;
  uint8_t *data; size_t len;
  tssh_event_t *E;
  tssh_buffer_t *rbuffer, dummy;
  uint64_t offset;

  retval= TSSH_FAIL;
  handle= NULL;
  retval= tssh_ftp_open_file(T, T->ftop_current, 0, SSH_FXF_READ, 0, &handle);
  if (retval) return retval;

  buffer= buffer_new(32*1024 + sizeof(dummy));
  buffer_write(buffer, &dummy, sizeof(dummy));
  offset= 0;
  while(1)
  {
    R= tssh_ftp_read_file(T, handle, &offset, 32*1024, &res);
    if (R==TSSH_EOF) break;
    if (R) goto fail;
    buffer_write(buffer, res->data, res->len);
    free(res);
  }

  retval= tssh_ftp_close_handle(T, handle);
  if (retval) goto fail_noclose;

  buffer_free(buffer, (void**) &rbuffer, &len);
  rbuffer->data= (char*) rbuffer + sizeof(*rbuffer);
  rbuffer->len= len - sizeof(*rbuffer);

  E= tssh_event_new(TSSH_EVENT_FTP_SUCCESS);
  E->ftp_op= T->ftop_current;
  T->ftop_current= NULL;
  E->buffer= rbuffer;
  tssh_report(T, E);
  return TSSH_OK;

fail:
  tssh_ftp_close_handle(T, handle);
fail_noclose:
  buffer_free(buffer, NULL, NULL);
  return retval;
}

static int tssh_ftp_do_downloadfile(tssh_t *T)
{
  int ofd;

  int retval;
  tssh_obj_t *handle;
  int R;
  tssh_buffer_t *res;
  tssh_event_t *E;
  uint64_t offset;

  ofd= open(T->ftop_current->localfn, O_WRONLY | O_TRUNC | O_CREAT, 0600);
  if (ofd<0) 
  {
    tssh_log(T, "open(%s): %s\n", T->ftop_current->localfn, strerror(errno));
    return TSSH_FAIL;
  }

  retval= TSSH_FAIL;
  handle= NULL;
  retval= tssh_ftp_open_file(T, T->ftop_current, 0, SSH_FXF_READ, 0, &handle);
  if (retval) return retval;

  offset= 0;
  while(1)
  {
    retval= tssh_ftp_read_file(T, handle, &offset, 32*1024, &res);
    if (retval==TSSH_EOF) break;
    if (retval) goto fail;
    while(res->len)
    {
      R= write(ofd, res->data, res->len);
      if (R<0) 
      { 
        if (errno==EINTR) continue;
        tssh_log(T, "write(): %s\n", strerror(errno));
        goto fail;
      }
      res->data += R;
      res->len -= R;
    }
    free(res);
  }

  retval= tssh_ftp_close_handle(T, handle);
  if (close(ofd)) 
  {
    tssh_log(T, "close(): %s\n", strerror(errno));
    goto fail_noclose;
  }
  if (retval) goto fail_noclose;


  E= tssh_event_new(TSSH_EVENT_FTP_SUCCESS);
  E->ftp_op= T->ftop_current;
  T->ftop_current= NULL;
  tssh_report(T, E);
  return TSSH_OK;

fail:
  tssh_ftp_close_handle(T, handle);
  close(ofd);
fail_noclose:
  return retval;
}

static int tssh_ftp_do_uploadfile(tssh_t *T)
{
  int ifd;

  int retval;
  tssh_obj_t *handle;
  int R;
  tssh_buffer_t *res;
  tssh_event_t *E;
  uint64_t offset;
  uint8_t *rbuffer;

  ifd= open(T->ftop_current->localfn, O_RDONLY);
  if (ifd<0) 
  {
    tssh_log(T, "open(%s): %s\n", T->ftop_current->localfn, strerror(errno));
    return TSSH_FAIL;
  }

  retval= TSSH_FAIL;
  handle= NULL;
  rbuffer= NULL;
  retval= tssh_ftp_open_file(T, T->ftop_current, 1, 
             SSH_FXF_WRITE | SSH_FXF_CREAT | SSH_FXF_TRUNC, 0600, &handle);
  if (retval) return retval;

  offset= 0;
  rbuffer= malloc(32*1024);
  while(1)
  {
    R= read(ifd, rbuffer, 32*1024);
    if (R==0) break;
    if (R<0)
    {
      if (errno==EINTR) continue;
      tssh_log(T, "read(%s): %s\n", T->ftop_current->localfn);
      goto fail;
    }
    retval= tssh_ftp_write_file(T, handle, &offset, rbuffer, R);
    if (retval) goto fail;
  }

  retval= tssh_ftp_close_handle(T, handle);
  if (close(ifd)) 
  {
    tssh_log(T, "close(): %s\n", strerror(errno));
    goto fail_noclose;
  }
  if (retval) goto fail_noclose;

  E= tssh_event_new(TSSH_EVENT_FTP_SUCCESS);
  E->ftp_op= T->ftop_current;
  T->ftop_current= NULL;
  tssh_report(T, E);
  return TSSH_OK;

fail:
  tssh_ftp_close_handle(T, handle);
  close(ifd);
fail_noclose:
  if (rbuffer) free(rbuffer);
  return retval;
}

static int tssh_ftp_do_uploadmem(tssh_t *T)
{
  int retval;
  tssh_obj_t *handle;
  int cp;
  uint64_t offset;
  uint8_t *udata;
  uint32_t ulen;

  retval= TSSH_FAIL;
  handle= NULL;
  retval= tssh_ftp_open_file(T, T->ftop_current, 1, 
             SSH_FXF_WRITE | SSH_FXF_CREAT | SSH_FXF_TRUNC, 0600, &handle);
  if (retval) return retval;

  offset= 0;
  udata= T->ftop_current->udata;
  ulen= T->ftop_current->size;
  while(ulen)
  {
    cp= 32*1024;
    if (cp>ulen) cp= ulen;
    retval= tssh_ftp_write_file(T, handle, &offset, udata, cp);
    if (retval) goto fail;
    udata += cp;
    ulen -= cp;
  }

  if (tssh_ftp_close_handle(T, handle))
    goto fail_noclose;

  return tssh_ftp_report_success_simple(T);

fail:
  tssh_ftp_close_handle(T, handle);
fail_noclose:
  return retval;
}

static int tssh_ftp_get_simple_cmd_response(tssh_t *T)
{
  int cmd;
  tssh_buffer_t *res;
  int R;
  tssh_event_t *E;
  if (tssh_ftp_wait_data(T, &cmd, &res, 1, SSH_FXP_STATUS)) 
    return TSSH_FAIL_HARD;
  R= tssh_ftp_report_status(T, res);
  free(res);
  if (R!=SSH_FX_OK) return TSSH_FAIL;
  E= tssh_event_new(TSSH_EVENT_FTP_SUCCESS);
  E->ftp_op= T->ftop_current;
  T->ftop_current= NULL;
  tssh_report(T, E);
  return TSSH_OK;
}

static int tssh_ftp_do_rmfile(tssh_t *T)
{
  if (tssh_ftp_cmd(T, "CRs", SSH_FXP_REMOVE, T->ftop_current->fn))
     return TSSH_FAIL_HARD;
  return tssh_ftp_get_simple_cmd_response(T);
}

static int tssh_ftp_do_rmdir(tssh_t *T)
{
  if (tssh_ftp_cmd(T, "CRs", SSH_FXP_RMDIR, T->ftop_current->fn))
     return TSSH_FAIL_HARD;
  return tssh_ftp_get_simple_cmd_response(T);
}

static int tssh_ftp_do_mkdir(tssh_t *T)
{
  if (tssh_ftp_cmd(T, "CRsdd", SSH_FXP_MKDIR, T->ftop_current->fn,
                          SSH_FILEXFER_ATTR_PERMISSIONS, 0700))
     return TSSH_FAIL_HARD;
  return tssh_ftp_get_simple_cmd_response(T);
}

static int tssh_ftp_do_rename(tssh_t *T)
{
  if (tssh_ftp_cmd(T, "CRss", SSH_FXP_RENAME, T->ftop_current->fn,
                             T->ftop_current->oldfn))
     return TSSH_FAIL_HARD;
  return tssh_ftp_get_simple_cmd_response(T);
}

static int tssh_ftp_do_query(tssh_t *T)
{
  uint32_t reqid;
  int cmd;
  tssh_buffer_t *res;
  tssh_fileinfo_t finfo;
  tssh_event_t *E;

  if (tssh_ftp_cmd(T, "CRs",SSH_FXP_STAT, T->ftop_current->fn))
    return TSSH_FAIL_HARD;
  if (tssh_ftp_wait_data(T, &cmd, &res, 2, SSH_FXP_ATTRS, SSH_FXP_STATUS))
    return TSSH_FAIL_HARD;

  if (cmd==SSH_FXP_STATUS)
  {
    tssh_ftp_report_status(T, res);
    free(res);
    return TSSH_FAIL;
  }

  if (tssh_xparse_uint32("f-reqid", res, &reqid) || 
      tssh_ftp_parse_file_attr(T, res, &finfo)) 
  {
    free(res);
    return TSSH_FAIL_HARD;
  }
  free(res);
  E= tssh_event_new(TSSH_EVENT_FTP_SUCCESS);
  E->ftp_op= T->ftop_current;
  T->ftop_current= NULL;
  E->fileinfo= malloc(sizeof(E->fileinfo[0]));
  E->fileinfo[0]= finfo;
  E->n_fileinfo= 1;
  tssh_report(T,E);
  return TSSH_OK;
}

